==> pubnub-8.4.1.tgz: package/src/transport/web-react-native-transport.ts <==
/**
 * Common browser and React Native Transport provider module.
 *
 * @internal
 */

import { CancellationController, TransportRequest } from '../core/types/transport-request';
import { TransportResponse } from '../core/types/transport-response';
import { PubNubAPIError } from '../errors/pubnub-api-error';
import { Transport } from '../core/interfaces/transport';
import { PubNubFileInterface } from '../core/types/file';
import { queryStringFromObject } from '../core/utils';

/**
 * Class representing a `fetch`-based browser and React Native transport provider.
 *
 * @internal
 */
export class WebReactNativeTransport implements Transport {
  /**
   * Service {@link ArrayBuffer} response decoder.
   *
   * @internal
   */
  protected static decoder = new TextDecoder();

  /**
   * Create and configure transport provider for Web and Rect environments.
   *
   * @param [keepAlive] - Whether client should try to keep connections open for reuse or not.
   * @param logVerbosity - Whether verbose logs should be printed or not.
   *
   * @internal
   */
  constructor(
    private keepAlive: boolean = false,
    private readonly logVerbosity: boolean,
  ) {}

  makeSendable(req: TransportRequest): [Promise<TransportResponse>, CancellationController | undefined] {
    let controller: CancellationController | undefined;
    let abortController: AbortController | undefined;

    if (req.cancellable) {
      abortController = new AbortController();
      controller = {
        // Storing controller inside to prolong object lifetime.
        abortController,
        abort: () => abortController?.abort(),
      } as CancellationController;
    }

    return [
      this.requestFromTransportRequest(req).then((request) => {
        const start = new Date().getTime();

        this.logRequestProcessProgress(request, req.body);

        /**
         * Setup request timeout promise.
         *
         * **Note:** Native Fetch API doesn't support `timeout` out-of-box.
         */
        const requestTimeout = new Promise<Response>((_, reject) => {
          const timeoutId = setTimeout(() => {
            // Clean up.
            clearTimeout(timeoutId);

            reject(new Error('Request timeout'));
          }, req.timeout * 1000);
        });

        return Promise.race([
          fetch(request, { signal: abortController?.signal, credentials: 'omit', cache: 'no-cache' }),
          requestTimeout,
        ])
          .then((response): Promise<[Response, ArrayBuffer]> | [Response, ArrayBuffer] =>
            response.arrayBuffer().then((arrayBuffer) => [response, arrayBuffer]),
          )
          .then((response) => {
            const responseBody = response[1].byteLength > 0 ? response[1] : undefined;
            const { status, headers: requestHeaders } = response[0];
            const headers: Record<string, string> = {};

            // Copy Headers object content into plain Record.
            requestHeaders.forEach((value, key) => (headers[key] = value.toLowerCase()));

            const transportResponse: TransportResponse = {
              status,
              url: request.url,
              headers,
              body: responseBody,
            };

            if (status >= 400) throw PubNubAPIError.create(transportResponse);

            this.logRequestProcessProgress(request, undefined, new Date().getTime() - start, responseBody);

            return transportResponse;
          })
          .catch((error) => {
            throw PubNubAPIError.create(error);
          });
      }),
      controller,
    ];
  }

  request(req: TransportRequest): TransportRequest {
    return req;
  }

  /**
   * Creates a Request object from a given {@link TransportRequest} object.
   *
   * @param req - The {@link TransportRequest} object containing request information.
   *
   * @returns Request object generated from the {@link TransportRequest} object.
   *
   * @internal
   */
  private async requestFromTransportRequest(req: TransportRequest): Promise<Request> {
    let body: string | ArrayBuffer | FormData | undefined;
    let path = req.path;

    // Create multipart request body.
    if (req.formData && req.formData.length > 0) {
      // Reset query parameters to conform to signed URL
      req.queryParameters = {};

      const file = req.body as PubNubFileInterface;
      const formData = new FormData();
      for (const { key, value } of req.formData) formData.append(key, value);
      try {
        const fileData = await file.toArrayBuffer();
        formData.append('file', new Blob([fileData], { type: 'application/octet-stream' }), file.name);
      } catch (_) {
        try {
          const fileData = await file.toFileUri();
          // @ts-expect-error React Native File Uri support.
          formData.append('file', fileData, file.name);
        } catch (_) {}
      }

      body = formData;
    }
    // Handle regular body payload (if passed).
    else if (req.body && (typeof req.body === 'string' || req.body instanceof ArrayBuffer)) body = req.body;

    if (req.queryParameters && Object.keys(req.queryParameters).length !== 0)
      path = `${path}?${queryStringFromObject(req.queryParameters)}`;

    return new Request(`${req.origin!}${path}`, {
      method: req.method,
      headers: req.headers,
      redirect: 'follow',
      body,
    });
  }

  /**
   * Log out request processing progress and result.
   *
   * @param request - Platform-specific
   * @param [requestBody] - POST / PATCH body.
   * @param [elapsed] - How many seconds passed since request processing started.
   * @param [body] - Service response (if available).
   *
   * @internal
   */
  protected logRequestProcessProgress(
    request: Request,
    requestBody: TransportRequest['body'],
    elapsed?: number,
    body?: ArrayBuffer,
  ) {
    if (!this.logVerbosity) return;

    const { protocol, host, pathname, search } = new URL(request.url);
    const timestamp = new Date().toISOString();

    if (!elapsed) {
      let outgoing = `[${timestamp}]\n${protocol}//${host}${pathname}\n${search}`;
      if (requestBody && (typeof requestBody === 'string' || requestBody instanceof ArrayBuffer)) {
        if (typeof requestBody === 'string') outgoing += `\n\n${requestBody}`;
        else outgoing += `\n\n${WebReactNativeTransport.decoder.decode(requestBody)}`;
      }

      console.log(`<<<<<`);
      console.log(outgoing);
      console.log('-----');
    } else {
      let outgoing = `[${timestamp} / ${elapsed}]\n${protocol}//${host}${pathname}\n${search}`;
      if (body) outgoing += `\n\n${WebReactNativeTransport.decoder.decode(body)}`;

      console.log('>>>>>>');
      console.log(outgoing);
      console.log('-----');
    }
  }
}

==> pubnub-8.7.1.tgz: package/src/transport/web-react-native-transport.ts <==
/**
 * Common browser and React Native Transport provider module.
 *
 * @internal
 */

import { CancellationController, TransportRequest } from '../core/types/transport-request';
import { TransportResponse } from '../core/types/transport-response';
import { PubNubAPIError } from '../errors/pubnub-api-error';
import { Transport } from '../core/interfaces/transport';
import { PubNubFileInterface } from '../core/types/file';
import { queryStringFromObject } from '../core/utils';

/**
 * Class representing a `fetch`-based browser and React Native transport provider.
 *
 * @internal
 */
export class WebReactNativeTransport implements Transport {
  /**
   * Pointer to the "clean" `fetch` function.
   *
   * This protects against APM which overload implementation and may break crucial functionality.
   *
   * @internal
   */
  private static originalFetch: typeof fetch;

  /**
   * Service {@link ArrayBuffer} response decoder.
   *
   * @internal
   */
  protected static decoder = new TextDecoder();

  /**
   * Create and configure transport provider for Web and Rect environments.
   *
   * @param originalFetch - Pointer to the original (not monkey patched) `fetch` implementation.
   * @param [keepAlive] - Whether client should try to keep connections open for reuse or not.
   * @param logVerbosity - Whether verbose logs should be printed or not.
   *
   * @internal
   */
  constructor(
    originalFetch: unknown,
    private keepAlive: boolean = false,
    private readonly logVerbosity: boolean = false,
  ) {
    WebReactNativeTransport.originalFetch = originalFetch as typeof fetch;

    // Check whether `fetch` has been monkey patched or not.
    if (logVerbosity && this.isFetchMonkeyPatched()) {
      console.warn("[PubNub] Native Web Fetch API 'fetch' function monkey patched.");
      if (!this.isFetchMonkeyPatched(WebReactNativeTransport.originalFetch))
        console.info("[PubNub] Use native Web Fetch API 'fetch' implementation from iframe as APM workaround.");
      else
        console.warn(
          '[PubNub] Unable receive native Web Fetch API. There can be issues with subscribe long-poll cancellation',
        );
    }
  }

  makeSendable(req: TransportRequest): [Promise<TransportResponse>, CancellationController | undefined] {
    let controller: CancellationController | undefined;
    let abortController: AbortController | undefined;

    if (req.cancellable) {
      abortController = new AbortController();
      controller = {
        // Storing controller inside to prolong object lifetime.
        abortController,
        abort: () => abortController?.abort(),
      } as CancellationController;
    }

    return [
      this.requestFromTransportRequest(req).then((request) => {
        const start = new Date().getTime();

        this.logRequestProcessProgress(request, req.body);

        /**
         * Setup request timeout promise.
         *
         * **Note:** Native Fetch API doesn't support `timeout` out-of-box.
         */
        const requestTimeout = new Promise<Response>((_, reject) => {
          const timeoutId = setTimeout(() => {
            // Clean up.
            clearTimeout(timeoutId);

            reject(new Error('Request timeout'));
          }, req.timeout * 1000);
        });

        return Promise.race([
          WebReactNativeTransport.originalFetch(request, {
            signal: abortController?.signal,
            credentials: 'omit',
            cache: 'no-cache',
          }),
          requestTimeout,
        ])
          .then((response): Promise<[Response, ArrayBuffer]> | [Response, ArrayBuffer] =>
            response.arrayBuffer().then((arrayBuffer) => [response, arrayBuffer]),
          )
          .then((response) => {
            const responseBody = response[1].byteLength > 0 ? response[1] : undefined;
            const { status, headers: requestHeaders } = response[0];
            const headers: Record<string, string> = {};

            // Copy Headers object content into plain Record.
            requestHeaders.forEach((value, key) => (headers[key] = value.toLowerCase()));

            const transportResponse: TransportResponse = {
              status,
              url: request.url,
              headers,
              body: responseBody,
            };

            if (status >= 400) throw PubNubAPIError.create(transportResponse);

            this.logRequestProcessProgress(request, undefined, new Date().getTime() - start, responseBody);

            return transportResponse;
          })
          .catch((error) => {
            throw PubNubAPIError.create(error);
          });
      }),
      controller,
    ];
  }

  request(req: TransportRequest): TransportRequest {
    return req;
  }

  /**
   * Creates a Request object from a given {@link TransportRequest} object.
   *
   * @param req - The {@link TransportRequest} object containing request information.
   *
   * @returns Request object generated from the {@link TransportRequest} object.
   *
   * @internal
   */
  private async requestFromTransportRequest(req: TransportRequest): Promise<Request> {
    let body: string | ArrayBuffer | FormData | undefined;
    let path = req.path;

    // Create multipart request body.
    if (req.formData && req.formData.length > 0) {
      // Reset query parameters to conform to signed URL
      req.queryParameters = {};

      const file = req.body as PubNubFileInterface;
      const formData = new FormData();
      for (const { key, value } of req.formData) formData.append(key, value);
      try {
        const fileData = await file.toArrayBuffer();
        formData.append('file', new Blob([fileData], { type: 'application/octet-stream' }), file.name);
      } catch (_) {
        try {
          const fileData = await file.toFileUri();
          // @ts-expect-error React Native File Uri support.
          formData.append('file', fileData, file.name);
        } catch (_) {}
      }

      body = formData;
    }
    // Handle regular body payload (if passed).
    else if (req.body && (typeof req.body === 'string' || req.body instanceof ArrayBuffer)) body = req.body;

    if (req.queryParameters && Object.keys(req.queryParameters).length !== 0)
      path = `${path}?${queryStringFromObject(req.queryParameters)}`;

    return new Request(`${req.origin!}${path}`, {
      method: req.method,
      headers: req.headers,
      redirect: 'follow',
      body,
    });
  }

  /**
   * Log out request processing progress and result.
   *
   * @param request - Platform-specific
   * @param [requestBody] - POST / PATCH body.
   * @param [elapsed] - How many seconds passed since request processing started.
   * @param [body] - Service response (if available).
   *
   * @internal
   */
  protected logRequestProcessProgress(
    request: Request,
    requestBody: TransportRequest['body'],
    elapsed?: number,
    body?: ArrayBuffer,
  ) {
    if (!this.logVerbosity) return;

    const { protocol, host, pathname, search } = new URL(request.url);
    const timestamp = new Date().toISOString();

    if (!elapsed) {
      let outgoing = `[${timestamp}]\n${protocol}//${host}${pathname}\n${search}`;
      if (requestBody && (typeof requestBody === 'string' || requestBody instanceof ArrayBuffer)) {
        if (typeof requestBody === 'string') outgoing += `\n\n${requestBody}`;
        else outgoing += `\n\n${WebReactNativeTransport.decoder.decode(requestBody)}`;
      }

      console.log(`<<<<<`);
      console.log(outgoing);
      console.log('-----');
    } else {
      let outgoing = `[${timestamp} / ${elapsed}]\n${protocol}//${host}${pathname}\n${search}`;
      if (body) outgoing += `\n\n${WebReactNativeTransport.decoder.decode(body)}`;

      console.log('>>>>>>');
      console.log(outgoing);
      console.log('-----');
    }
  }

  /**
   * Check whether original `fetch` has been monkey patched or not.
   *
   * @returns `true` if original `fetch` has been patched.
   *
   * @internal
   */
  private isFetchMonkeyPatched(oFetch?: typeof fetch): boolean {
    const fetchString = (oFetch ?? fetch).toString();

    return !fetchString.includes('[native code]') && fetch.name !== 'fetch';
  }
}
